@@ -43,6 +43,42 @@ test_vector_expression_literal (void)
 }
 
 
+static void
+test_vector_expression_number_array (void)
+{
+  g_autoptr(GError) error = NULL;
+  g_autoptr(JsonNode) node = json_from_string ("[1, 2, 3, 4, 5]", NULL);
+  g_autoptr(ShumateVectorExpression) expression;
+  g_auto(ShumateVectorValue) value = SHUMATE_VECTOR_VALUE_INIT;
+  double num;
+
+  expression = shumate_vector_expression_from_json (node, &error);
+  g_assert_no_error (error);
+
+  g_assert_true (shumate_vector_expression_eval (expression, NULL, &value));
+  g_assert_cmpint (5, ==, value.array->len);
+  for (int i = 0; i < 5; i++)
+    {
+      g_assert_true (shumate_vector_value_get_number (value.array->pdata[i], &num));
+      g_assert_cmpfloat (i + 1, ==, num);
+    }
+}
+
+
+static void
+test_vector_expression_nested_array_literal (void)
+{
+  g_autoptr(GError) error = NULL;
+  g_autoptr(JsonNode) node = json_from_string ("[\"in\", 2, [1, 2, 3, 4, 5]]", NULL);
+  g_autoptr(ShumateVectorExpression) expression;
+  g_auto(ShumateVectorValue) value = SHUMATE_VECTOR_VALUE_INIT;
+  double num;
+
+  expression = shumate_vector_expression_from_json (node, &error);
+  g_assert_error (error, SHUMATE_STYLE_ERROR, SHUMATE_STYLE_ERROR_INVALID_EXPRESSION);
+}
+
+
 static void
 check_interpolate (ShumateVectorExpression *expression)
 {
@@ -290,11 +326,68 @@ test_vector_expression_basic_filter (void)
   g_assert_true (filter ("[\"==\", [\"slice\", \"Hello, \U0001F30E!\", 7], \"\U0001F30E!\"]"));
   g_assert_true (filter ("[\"==\", [\"slice\", \"Hello, \U0001F30E!\", 7, 8], \"\U0001F30E\"]"));
 
+  /* Test slice with negative and out of range indices */
+  g_assert_true (filter ("[\"==\", [\"slice\", [\"literal\", [\"a\", \"b\", \"c\"]], -2], [\"literal\", [\"b\", \"c\"]]]"));
+  g_assert_true (filter ("[\"==\", [\"slice\", [\"literal\", [\"a\", \"b\", \"c\"]], -3], [\"literal\", [\"a\", \"b\", \"c\"]]]"));
+  g_assert_true (filter ("[\"==\", [\"slice\", [\"literal\", [\"a\", \"b\", \"c\"]], -4], [\"literal\", [\"a\", \"b\", \"c\"]]]"));
+  g_assert_true (filter ("[\"==\", [\"slice\", [\"literal\", [\"a\", \"b\", \"c\"]], 0, -1], [\"literal\", [\"a\", \"b\"]]]"));
+  g_assert_true (filter ("[\"==\", [\"slice\", [\"literal\", [\"a\", \"b\", \"c\"]], 0, -3], [\"literal\", []]]"));
+  g_assert_true (filter ("[\"==\", [\"slice\", [\"literal\", [\"a\", \"b\", \"c\"]], 0, -4], [\"literal\", []]]"));
+  g_assert_true (filter ("[\"==\", [\"slice\", [\"literal\", [\"a\", \"b\", \"c\"]], 3, 4], [\"literal\", []]]"));
+  g_assert_true (filter ("[\"==\", [\"slice\", \"abc\", -2], \"bc\"]"));
+  g_assert_true (filter ("[\"==\", [\"slice\", \"abc\", -3], \"abc\"]"));
+  g_assert_true (filter ("[\"==\", [\"slice\", \"abc\", -4], \"abc\"]"));
+  g_assert_true (filter ("[\"==\", [\"slice\", \"abc\", 0, -1], \"ab\"]"));
+  g_assert_true (filter ("[\"==\", [\"slice\", \"abc\", 0, -3], \"\"]"));
+  g_assert_true (filter ("[\"==\", [\"slice\", \"abc\", 0, -4], \"\"]"));
+  g_assert_true (filter ("[\"==\", [\"slice\", \"abc\", 3, 4], \"\"]"));
+
   g_assert_true (filter ("[\"==\", [\"literal\", \"hello\"], \"HELLO\", [\"collator\", {\"case-sensitive\": false}]]"));
   g_assert_true (filter ("[\"!=\", [\"literal\", \"hello\"], \"HELLO\", [\"collator\", {\"case-sensitive\": true}]]"));
   g_assert_true (filter ("[\">\", [\"literal\", \"hello\"], \"a\", [\"collator\", {}]]"));
   g_assert_true (filter ("[\"<\", [\"literal\", \"a\"], \"hello\", [\"collator\", {}]]"));
   g_assert_true (filter ("[\"!=\", [\"resolved-locale\", [\"collator\", {}]], \"foo\"]"));
+
+  g_assert_true (filter ("[\"==\", [\"step\", -10, \"a\", 1, \"b\", 2, \"c\"], \"a\"]"));
+  g_assert_true (filter ("[\"==\", [\"step\", 1.5, \"a\", 1, \"b\", 2, \"c\"], \"b\"]"));
+  g_assert_true (filter ("[\"==\", [\"step\", 2, \"a\", 1, \"b\", 2, \"c\"], \"c\"]"));
+  g_assert_true (filter ("[\"==\", [\"step\", 3, \"a\", 1, \"b\", 2, \"c\"], \"c\"]"));
+
+  g_assert_true (filter ("[\"==\", [\"to-boolean\", 0], false]"));
+  g_assert_true (filter ("[\"==\", [\"to-boolean\", 1], true]"));
+  g_assert_true (filter ("[\"==\", [\"to-boolean\", -2], true]"));
+  g_assert_true (filter ("[\"==\", [\"to-boolean\", null], false]"));
+  g_assert_true (filter ("[\"==\", [\"to-boolean\", \"\"], false]"));
+  g_assert_true (filter ("[\"==\", [\"to-boolean\", \"hello\"], true]"));
+  g_assert_true (filter ("[\"==\", [\"to-boolean\", [\"collator\", {}]], true]"));
+  g_assert_true (filter ("[\"==\", [\"to-color\", \"red\"], [\"to-color\", \"rgb(255, 0, 0)\"]]"));
+  g_assert_true (filter ("[\"==\", [\"to-color\", \"not a color\", \"#FF0000\"], [\"to-color\", \"rgb(255, 0, 0)\"]]"));
+  g_assert_true (filter ("[\"==\", [\"to-number\", \"2\"], 2]"));
+  g_assert_true (filter ("[\"==\", [\"to-number\", \"-.5\"], -0.5]"));
+  g_assert_true (filter ("[\"==\", [\"to-number\", \"1e3\"], 1000]"));
+  g_assert_true (filter ("[\"==\", [\"to-number\", null], 0]"));
+  g_assert_true (filter ("[\"==\", [\"to-number\", false], 0]"));
+  g_assert_true (filter ("[\"==\", [\"to-number\", true], 1]"));
+  g_assert_true (filter ("[\"==\", [\"to-number\", \"not a number\", 10], 10]"));
+  g_assert_false (filter ("[\"==\", 2, \"2\"]"));
+  g_assert_true (filter ("[\"==\", [\"to-string\", 2], \"2\"]"));
+  g_assert_true (filter ("[\"==\", [\"to-string\", \"a\"], \"a\"]"));
+  g_assert_true (filter ("[\"==\", [\"to-string\", true], \"true\"]"));
+  g_assert_true (filter ("[\"==\", [\"to-string\", false], \"false\"]"));
+  g_assert_true (filter ("[\"==\", [\"to-string\", null], \"\"]"));
+  g_assert_true (filter ("[\"==\", [\"to-string\", [\"to-color\", \"gold\"]], \"rgba(255,215,0,1)\"]"));
+  g_assert_true (filter ("[\"==\", [\"to-string\", [\"to-color\", \"rgba(255, 1, 2, 0.1)\"]], \"rgba(255,1,2,0.1)\"]"));
+  g_assert_true (filter ("[\"==\", [\"to-string\", [\"literal\", [1, 0.5, null, true, [\"b\"]]]], \"[1,0.5,null,true,[\\\"b\\\"]]\"]"));
+
+  /* Test NaN/inf handling */
+  g_assert_true (filter ("[\"==\", [\"to-boolean\", [\"/\", 0, 0]], false]"));
+  g_assert_true (filter ("[\"==\", [\"to-boolean\", [\"/\", 1, 0]], true]"));
+  g_assert_true (filter ("[\"==\", [\"to-string\", [\"/\", 0, 0]], \"NaN\"]"));
+  g_assert_true (filter ("[\"==\", [\"to-string\", [\"/\", 1, 0]], \"Infinity\"]"));
+  g_assert_true (filter ("[\"==\", [\"to-string\", [\"/\", -1, 0]], \"-Infinity\"]"));
+  g_assert_true (filter ("[\"==\", [\"to-string\", [\"%\", 0, 0]], \"NaN\"]"));
+  g_assert_true (filter ("[\"==\", [\"to-string\", [\"%\", 1, 0]], \"NaN\"]"));
+  g_assert_true (filter ("[\"==\", [\"to-string\", [\"%\", -1, 0]], \"NaN\"]"));
 }
 
 
@@ -338,21 +431,20 @@ test_vector_expression_feature_filter (void)
 {
   GError *error = NULL;
   g_autoptr(GBytes) vector_data = NULL;
-  gconstpointer data;
-  gsize len;
+  g_autoptr(ShumateVectorReader) reader = NULL;
   ShumateVectorRenderScope scope;
 
   vector_data = g_resources_lookup_data ("/org/gnome/shumate/Tests/0.pbf", G_RESOURCE_LOOKUP_FLAGS_NONE, NULL);
   g_assert_no_error (error);
 
-  data = g_bytes_get_data (vector_data, &len);
-  scope.tile = vector_tile__tile__unpack (NULL, len, data);
-  g_assert_nonnull (scope.tile);
+  reader = shumate_vector_reader_new (vector_data);
+  scope.reader = shumate_vector_reader_iterate (reader);
+  g_assert_nonnull (scope.reader);
 
   scope.zoom_level = 10;
 
-  g_assert_true (shumate_vector_render_scope_find_layer (&scope, "helloworld"));
-  scope.feature = scope.layer->features[0];
+  g_assert_true (shumate_vector_reader_iter_read_layer_by_name (scope.reader, "helloworld"));
+  g_assert_true (shumate_vector_reader_iter_next_feature (scope.reader));
 
   g_assert_true  (filter_with_scope (&scope, "[\"==\", \"name\", \"Hello, world!\"]"));
   g_assert_true  (filter_with_scope (&scope, "[\"==\", [\"get\", \"name\"], \"Hello, world!\"]"));
@@ -379,7 +471,7 @@ test_vector_expression_feature_filter (void)
   g_assert_true  (filter_with_scope (&scope, "[\"==\", [\"concat\", \"Hello, world!\"], \"Hello, world!\"]"));
   g_assert_true  (filter_with_scope (&scope, "[\"!=\", [\"concat\", \"Hello, world!\"], \"HELLO, WORLD!\"]"));
 
-  vector_tile__tile__free_unpacked (scope.tile, NULL);
+  g_clear_object (&scope.reader);
 }
 
 
@@ -416,8 +508,7 @@ test_vector_expression_format ()
 {
   GError *error = NULL;
   g_autoptr(GBytes) vector_data = NULL;
-  gconstpointer data;
-  gsize len;
+  g_autoptr(ShumateVectorReader) reader = NULL;
   ShumateVectorRenderScope scope;
   g_autoptr(JsonNode) node = json_from_string ("\"***** {name} *****\"", NULL);
   g_autoptr(ShumateVectorExpression) expression;
@@ -429,19 +520,19 @@ test_vector_expression_format ()
   vector_data = g_resources_lookup_data ("/org/gnome/shumate/Tests/0.pbf", G_RESOURCE_LOOKUP_FLAGS_NONE, NULL);
   g_assert_no_error (error);
 
-  data = g_bytes_get_data (vector_data, &len);
-  scope.tile = vector_tile__tile__unpack (NULL, len, data);
-  g_assert_nonnull (scope.tile);
+  reader = shumate_vector_reader_new (vector_data);
+  scope.reader = shumate_vector_reader_iterate (reader);
+  g_assert_nonnull (scope.reader);
 
   scope.zoom_level = 10;
 
-  g_assert_true (shumate_vector_render_scope_find_layer (&scope, "helloworld"));
-  scope.feature = scope.layer->features[0];
+  g_assert_true (shumate_vector_reader_iter_read_layer_by_name (scope.reader, "helloworld"));
+  g_assert_true (shumate_vector_reader_iter_next_feature (scope.reader));
 
   result = shumate_vector_expression_eval_string (expression, &scope, NULL);
   g_assert_cmpstr (result, ==, "***** Hello, world! *****");
 
-  vector_tile__tile__free_unpacked (scope.tile, NULL);
+  g_clear_object (&scope.reader);
 }
 
 
@@ -467,7 +558,7 @@ test_vector_expression_array ()
   shumate_vector_value_array_append (&array1, &element2);
 
   string = shumate_vector_value_as_string (&array1);
-  g_assert_cmpstr (string, ==, "[Hello, world!, true]");
+  g_assert_cmpstr (string, ==, "[\"Hello, world!\",true]");
 
   shumate_vector_value_start_array (&array2);
   shumate_vector_value_array_append (&array2, &element1);
@@ -550,6 +641,8 @@ main (int argc, char *argv[])
 
   g_test_add_func ("/vector/expression/parse", test_vector_expression_parse);
   g_test_add_func ("/vector/expression/literal", test_vector_expression_literal);
+  g_test_add_func ("/vector/expression/number-array", test_vector_expression_number_array);
+  g_test_add_func ("/vector/expression/nested-array-literal", test_vector_expression_nested_array_literal);
   g_test_add_func ("/vector/expression/interpolate", test_vector_expression_interpolate);
   g_test_add_func ("/vector/expression/interpolate-filter", test_vector_expression_interpolate_filter);
   g_test_add_func ("/vector/expression/interpolate-color", test_vector_expression_interpolate_color);
